#include <assert.h>
#include <malloc.h>
#include "sprite.h"
#include "bitmask.h"
#include "refcount.h"

struct sprite_global sprite_global = 
{
  /* .display */ 0,
  /* .xoff = */ 0,
  /*.yoff =*/ 0,
  /*.bg_image = */0,
  /*.bg_mask =*/ 0,
  /*.dt =*/ 30,
  /*.display_dt = */30,
  /*.frame_start =*/ 0,
  /*.game_clock =*/ 0,
  /*.opt_alpha =*/ 2,
};

static int in_frame = 0; /* this is 1 if we are inside a frame */


void sprite_timer_set(sprite_timer_t *timer, unsigned int delay)
{
  *timer = delay + sprite_global.game_clock;
}

int sprite_timer_finished(sprite_timer_t timer)
{
  return timer <= sprite_global.game_clock;
}

static int nr_alarms = 0;
static int max_alarms = 32; /* initial max */
static struct alarm
{
  sprite_t *target;
  int signal;
  void *data;
  unsigned int at;
} *alarms;

void sprite_alarm(unsigned int delay, sprite_t *target, int signal, void *data)
{
  if (sprite_alive(target))
    {
      if (nr_alarms >= max_alarms)
	{
	  max_alarms*=2;
	  alarms = realloc(alarms,max_alarms*sizeof(*alarms));
	  assert(alarms);
	}
      alarms[nr_alarms].at = sprite_global.game_clock + delay;
      alarms[nr_alarms].target = target;
      alarms[nr_alarms].signal = signal;
      alarms[nr_alarms].data = data;
      REF(target);
      nr_alarms++;
    }
}

void sprite_start_frame()
{
  int i;
  in_frame = 1;
  sprite_global.frame_start = SDL_GetTicks();
  sprite_global.game_clock += sprite_global.dt;
  sprite_background_draw();
  for (i=0;i<nr_alarms;i++)
    {
      if (!sprite_alive(alarms[i].target))
	{
	  DEREF(alarms[i].target,alarms[i].target->type->free_sprite);
	  alarms[i--] = alarms[--nr_alarms];
	}
      else if (alarms[i].at <= sprite_global.game_clock)
	{
	  sprite_signal(alarms[i].target,alarms[i].signal,alarms[i].data);
	  DEREF(alarms[i].target,alarms[i].target->type->free_sprite);
	  alarms[i--] = alarms[--nr_alarms];
	}
    }
}

static void sprite_dirty_update();

unsigned int sprite_end_frame()
{
  unsigned int dt;  
  sprite_dirty_update();
  in_frame = 0;
  dt = SDL_GetTicks() - sprite_global.frame_start;
  if (dt < sprite_global.display_dt)
    {
      SDL_Delay(sprite_global.display_dt - dt);
      /* for some reason SDL_Delay might sleep up to 10 ms to little,
	 and hence the stuff below (Linux 2.4.18, debian) */
      dt = SDL_GetTicks() - sprite_global.frame_start;
      if (dt < sprite_global.display_dt - 5)
	{
	  SDL_Delay(sprite_global.display_dt - dt);
	}
      return sprite_global.display_dt - dt;
    }
  else return 0;
}

/* Dirty tile map algo */
#define DIRTY_TILE_SIZE 64 /* This implies a maximum resolution of 2048 */
static int all_dirty = 1;
static int bg_all_dirty = 1;
static unsigned long int dirty_mask[32];
static unsigned long int bg_dirty_mask[32];
static SDL_Rect tile_rects[256];
static int nr_tile_rects;
static int nr_dirty;

static INLINE void dirty_rect(SDL_Rect *r)
{
  int xs,xe,ys,ye,i,j;
  xs = r->x / DIRTY_TILE_SIZE;
  ys = r->y / DIRTY_TILE_SIZE;  
  xe = (r->x + r->w - 1) / DIRTY_TILE_SIZE;
  ye = (r->y + r->h - 1) / DIRTY_TILE_SIZE;
  for (i=ys;i<=ye;i++)
    for (j = xs; j <= xe; j++)
      dirty_mask[i] |= (unsigned long)1 << j;
}

void sprite_dirty(SDL_Rect *r)
{
  if ((!all_dirty) && (r->w > 0) && (r->h > 0) 
      && (r->x < sprite_global.display->w)
      && (r->y < sprite_global.display->h))
    {
      dirty_rect(r);
    }
}
 
void sprite_dirty_all()
{
  bg_all_dirty = 1;
  all_dirty = 1;
}

static INLINE void update_row(int row)
{
  SDL_Rect r;
  unsigned long int mask = dirty_mask[row];

  r.y = row*DIRTY_TILE_SIZE;
  if ((row + 1)*DIRTY_TILE_SIZE <= sprite_global.display->h)
    {
      r.h = DIRTY_TILE_SIZE;
    }
  else
    {
      r.h = DIRTY_TILE_SIZE - ((row + 1)*DIRTY_TILE_SIZE - sprite_global.display->h);
    }
  r.x = 0;
  r.w = 0;
  
  while (mask)
    {
      while (mask && ((mask & 1) == 0))
	{
	  r.x += DIRTY_TILE_SIZE;
	  mask /= 2;
	}
      while (mask && (mask & 1))
	{
	  r.w += DIRTY_TILE_SIZE;
	  mask /= 2;
	}
      tile_rects[nr_tile_rects++] = r;
      r.x += r.w;
      r.w = 0;
    }
  if (nr_tile_rects)
    {
      if (tile_rects[nr_tile_rects - 1].x + 
	  tile_rects[nr_tile_rects - 1].w > sprite_global.display->w)	  
      {
	tile_rects[nr_tile_rects - 1].w -= 
	  tile_rects[nr_tile_rects - 1].x + 
	  tile_rects[nr_tile_rects - 1].w - sprite_global.display->w;
      }
    }
}

/* Makes sure the dirty rects are visible 
 * and schedule them for background update next
 * frame. 
 */
static void sprite_dirty_update() 
{
  int i;
  unsigned long int mask;
  if (all_dirty)
    {
      SDL_UpdateRect(sprite_global.display,0,0,0,0);
      bg_all_dirty = 1;
    }
  else
    {
      for (i=0;i<32;i++)
	{
	  mask = dirty_mask[i];
	  dirty_mask[i] |= bg_dirty_mask[i]; /* Update also last frames tiles*/
	  bg_dirty_mask[i] = mask;
	  update_row(i);
	}
      SDL_UpdateRects(sprite_global.display,nr_tile_rects,tile_rects);
    }
  nr_tile_rects = 0;
  all_dirty = 0;
  for (i=0;i<32;i++)
    {      
      dirty_mask[i] = 0;
    }
}

void sprite_background_resize(int xsize, int ysize)
{
  SDL_Surface *surf;

  if (sprite_global.bg_image)
    sprite_background_free();
  sprite_global.bg_mask = bitmask_create(xsize,ysize);
  surf = SDL_CreateRGBSurface(SDL_SWSURFACE,xsize,ysize,
			      32,0xff000000,0xff0000,0xff00,0x0);
  assert(surf);
  sprite_global.bg_image = SDL_DisplayFormat(surf);
  SDL_FreeSurface(surf);
}

void sprite_background_free()
{
  SDL_FreeSurface(sprite_global.bg_image);
  bitmask_free(sprite_global.bg_mask);
  sprite_global.bg_image = 0;
  sprite_global.bg_mask = 0;
}

void sprite_background_compose(sprite_t *brush, int use_mask)
{
  SDL_Rect r;
  r.x = brush->x - brush->animation->xoff;
  r.y = brush->y - brush->animation->yoff;
  r.w = brush->animation->images[brush->anim_p]->w;
  r.h = brush->animation->images[brush->anim_p]->h;
  SDL_BlitSurface(brush->animation->images[brush->anim_p],0,
		  sprite_global.bg_image,&r);
  if (use_mask)
    {
      bitmask_draw(sprite_global.bg_mask, 
		   brush->animation->masks[brush->anim_p],
		   brush->x - brush->animation->xoff,
		   brush->y - brush->animation->yoff);
    }
}

static INLINE void background_update_row(int row)
{
  SDL_Rect r,rb;
  unsigned long int mask = bg_dirty_mask[row];

  r.y = row*DIRTY_TILE_SIZE;
  r.h = DIRTY_TILE_SIZE;
  r.x = 0;
  r.w = 0;
  
  while (mask)
    {
      while (mask && ((mask & 1) == 0))
	{
	  r.x += DIRTY_TILE_SIZE;
	  mask /= 2;
	}
      while (mask && (mask & 1))
	{
	  r.w += DIRTY_TILE_SIZE;
	  mask /= 2;
	}
      rb.x = r.x + sprite_global.xoff;
      rb.y = r.y + sprite_global.yoff;
      rb.w = r.w;
      rb.h = r.h;
      SDL_BlitSurface(sprite_global.bg_image,&rb,sprite_global.display,&r);
      r.x += r.w;
      r.w = 0;
    }
}

void sprite_background_draw()
{
  SDL_Rect r;
  int i;
  if (bg_all_dirty)
    {
      r.x = sprite_global.xoff;
      r.y = sprite_global.yoff;
      r.w = sprite_global.display->w;
      r.h = sprite_global.display->h;
      SDL_BlitSurface(sprite_global.bg_image,&r,sprite_global.display,0);
      bg_all_dirty = 0;
    }
  else
    {
      for (i=0;i<32;i++)
	background_update_row(i);
    }
}

void sprite_setup()
{
  alarms = malloc(sizeof(*alarms)*max_alarms);
}
